;; Generated package description from gnosis.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "gnosis" "0.10.4" "Knowledge System" '((emacs "29.1") (compat "29.1.4.2") (keymap-popup "0.2.0")) :commit "952a718b0ae7c1af0992cbc5aa80e1184999aa8a" :authors '(("Thanos Apollo" . "public@thanosapollo.org")) :maintainer '("Thanos Apollo" . "public@thanosapollo.org") :keywords '("extensions") :url "https://thanosapollo.org/projects/gnosis")
